from dataclasses import dataclass
from datetime import timedelta
from typing import Self

from eq3btsmart.models.base_model import BaseModel
from eq3btsmart.structures import PresetsStruct


@dataclass
class Presets(BaseModel[PresetsStruct]):
    window_open_temperature: float
    window_open_time: timedelta
    comfort_temperature: float
    eco_temperature: float
    offset_temperature: float

    @classmethod
    def from_struct(cls, struct: PresetsStruct) -> Self:
        return cls(
            window_open_temperature=struct.window_open_temp,
            window_open_time=struct.window_open_time,
            comfort_temperature=struct.comfort_temp,
            eco_temperature=struct.eco_temp,
            offset_temperature=struct.offset,
        )

    @classmethod
    def struct_type(cls) -> type[PresetsStruct]:
        return PresetsStruct



from datetime import timedelta

from eq3btsmart.models.presets import Presets
from eq3btsmart.structures import PresetsStruct


def test_presets_from_struct() -> None:
    preset_struct = PresetsStruct(
        window_open_temp=4.5,
        window_open_time=timedelta(seconds=5),
        comfort_temp=24,
        eco_temp=16,
        offset=0,
    )

    presets = Presets.from_struct(preset_struct)

    assert presets.window_open_temperature == 4.5
    assert presets.window_open_time == timedelta(seconds=5)
    assert presets.comfort_temperature == 24
    assert presets.eco_temperature == 16
    assert presets.offset_temperature == 0


def test_presets_struct_type() -> None:
    assert Presets.struct_type() == PresetsStruct



import pytest

from eq3btsmart.adapter.eq3_temperature_offset import Eq3TemperatureOffset
from eq3btsmart.const import EQ3BT_MAX_OFFSET, EQ3BT_MIN_OFFSET
from eq3btsmart.exceptions import Eq3InvalidDataException


def test_encode_valid_temperature_offset() -> None:
    offset = 1.0
    encoded = Eq3TemperatureOffset.encode(offset)
    assert encoded == 9


def test_decode_valid_int() -> None:
    encoded_int = 9
    decoded = Eq3TemperatureOffset.decode(encoded_int)
    assert decoded == 1.0


def test_round_trip_consistency() -> None:
    offset = 1.0
    encoded = Eq3TemperatureOffset.encode(offset)
    decoded = Eq3TemperatureOffset.decode(encoded)
    assert decoded == offset


def test_encode_temperature_offset_out_of_range_low() -> None:
    offset = EQ3BT_MIN_OFFSET - 0.1
    with pytest.raises(Eq3InvalidDataException):
        Eq3TemperatureOffset.encode(offset)


def test_encode_temperature_offset_out_of_range_high() -> None:
    offset = EQ3BT_MAX_OFFSET + 0.1
    with pytest.raises(Eq3InvalidDataException):
        Eq3TemperatureOffset.encode(offset)



from eq3btsmart.models.base_model import BaseModel as BaseModel
from eq3btsmart.models.device_data import DeviceData as DeviceData
from eq3btsmart.models.presets import Presets as Presets
from eq3btsmart.models.schedule import Schedule as Schedule
from eq3btsmart.models.schedule_day import ScheduleDay as ScheduleDay
from eq3btsmart.models.schedule_hour import ScheduleHour as ScheduleHour
from eq3btsmart.models.status import Status as Status



from dataclasses import dataclass, field
from typing import Self

from eq3btsmart.const import WeekDay
from eq3btsmart.models.schedule_day import ScheduleDay


@dataclass
class Schedule:
    schedule_days: list[ScheduleDay] = field(default_factory=list)

    def merge(self, other_schedule: Self) -> None:
        for other_schedule_day in other_schedule.schedule_days:
            schedule_day = next(
                (
                    schedule_day
                    for schedule_day in self.schedule_days
                    if schedule_day.week_day == other_schedule_day.week_day
                ),
                None,
            )

            if not schedule_day:
                self.schedule_days.append(other_schedule_day)
                continue

            schedule_day.schedule_hours = other_schedule_day.schedule_hours

    def delete_day(self, week_day: WeekDay) -> None:
        self.schedule_days = [
            schedule_day
            for schedule_day in self.schedule_days
            if schedule_day.week_day != week_day
        ]

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        return cls(schedule_days=[ScheduleDay.from_bytes(data)])

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, Schedule):
            return False

        week_days_to_compare = [
            schedule_day.week_day
            for schedule_day in self.schedule_days
            if len(schedule_day.schedule_hours) > 0
        ]
        week_days_to_compare.extend(
            [
                schedule_day.week_day
                for schedule_day in __value.schedule_days
                if len(schedule_day.schedule_hours) > 0
            ]
        )

        for week_day in week_days_to_compare:
            schedule_day = next(
                (
                    schedule_day
                    for schedule_day in self.schedule_days
                    if schedule_day.week_day == week_day
                ),
                None,
            )
            other_schedule_day = next(
                (
                    schedule_day
                    for schedule_day in __value.schedule_days
                    if schedule_day.week_day == week_day
                ),
                None,
            )

            if schedule_day is None or other_schedule_day is None:
                return False

            other_schedule_day = next(
                (
                    other_schedule_day
                    for other_schedule_day in __value.schedule_days
                    if other_schedule_day.week_day == schedule_day.week_day
                ),
                None,
            )

            if schedule_day != other_schedule_day:
                return False

        return True



from datetime import time

from eq3btsmart.const import WeekDay
from eq3btsmart.models import ScheduleDay, ScheduleHour
from eq3btsmart.structures import ScheduleDayStruct, ScheduleHourStruct


def test_schedule_day_initialization() -> None:
    schedule_day = ScheduleDay(
        week_day=WeekDay.MONDAY,
        schedule_hours=[ScheduleHour(21.5, time(hour=22, minute=0))],
    )
    assert schedule_day.week_day == WeekDay.MONDAY
    assert len(schedule_day.schedule_hours) == 1
    assert schedule_day.schedule_hours[0].target_temperature == 21.5


def test_schedule_day_from_struct() -> None:
    struct = ScheduleDayStruct(
        day=WeekDay.TUESDAY,
        hours=[
            ScheduleHourStruct(
                next_change_at=time(hour=22, minute=0),
                target_temp=22,
            )
        ],
    )

    schedule_day = ScheduleDay.from_struct(struct)

    assert schedule_day.week_day == WeekDay.TUESDAY
    assert len(schedule_day.schedule_hours) == 1
    assert schedule_day.schedule_hours[0].target_temperature == 22


def test_schedule_day_equality() -> None:
    day1 = ScheduleDay(
        week_day=WeekDay.WEDNESDAY,
        schedule_hours=[ScheduleHour(20.0, time(hour=22, minute=0))],
    )
    day2 = ScheduleDay(
        week_day=WeekDay.WEDNESDAY,
        schedule_hours=[ScheduleHour(20.0, time(hour=22, minute=0))],
    )
    day3 = ScheduleDay(
        week_day=WeekDay.THURSDAY,
        schedule_hours=[],
    )
    day4 = ScheduleDay(
        week_day=WeekDay.WEDNESDAY,
        schedule_hours=[],
    )

    assert day1 == day2
    assert day1 != day3
    assert day1 != day4
    assert day1 != "not a schedule day"



from datetime import datetime

from construct_typed import Adapter, Context

from eq3btsmart.exceptions import Eq3InvalidDataException


class Eq3Time(Adapter[bytes, bytes, datetime, datetime]):
    """Adapter to encode and decode time data."""

    def _encode(self, obj: datetime, ctx: Context, path: str) -> bytes:
        return self.encode(obj)

    def _decode(self, obj: bytes, ctx: Context, path: str) -> datetime:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: datetime) -> bytes:
        return bytes(
            [
                value.year % 100,
                value.month,
                value.day,
                value.hour,
                value.minute,
                value.second,
            ]
        )

    @classmethod
    def decode(cls, value: bytes) -> datetime:
        try:
            (year, month, day, hour, minute, second) = value
        except ValueError:
            raise Eq3InvalidDataException("Invalid time data")

        year += 2000

        return datetime(
            year=year, month=month, day=day, hour=hour, minute=minute, second=second
        )



from datetime import time

from construct_typed import Adapter, Context


class Eq3ScheduleTime(Adapter[int, int, time, time]):
    """Adapter to encode and decode schedule time data."""

    def _encode(self, obj: time, ctx: Context, path: str) -> int:
        return self.encode(obj)

    def _decode(self, obj: int, ctx: Context, path: str) -> time:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: time) -> int:
        return int((value.hour * 60 + value.minute) / 10)

    @classmethod
    def decode(cls, value: int) -> time:
        hour, minute = divmod(value * 10, 60)
        return time(hour=hour, minute=minute)



import pytest

from eq3btsmart.adapter.eq3_temperature import Eq3Temperature
from eq3btsmart.const import EQ3BT_OFF_TEMP, EQ3BT_ON_TEMP
from eq3btsmart.exceptions import Eq3InvalidDataException


def test_encode_valid_temperature() -> None:
    temperature = 22.0
    encoded = Eq3Temperature.encode(temperature)
    assert encoded == 44


def test_decode_valid_int() -> None:
    encoded_int = 44
    decoded = Eq3Temperature.decode(encoded_int)
    assert decoded == 22.0


def test_round_trip_consistency() -> None:
    temperature = 22.0
    encoded = Eq3Temperature.encode(temperature)
    decoded = Eq3Temperature.decode(encoded)
    assert decoded == temperature


def test_encode_temperature_out_of_range_low() -> None:
    temperature = EQ3BT_OFF_TEMP - 0.1
    with pytest.raises(Eq3InvalidDataException):
        Eq3Temperature.encode(temperature)


def test_encode_temperature_out_of_range_high() -> None:
    temperature = EQ3BT_ON_TEMP + 0.1
    with pytest.raises(Eq3InvalidDataException):
        Eq3Temperature.encode(temperature)



from construct_typed import Adapter, Context

from eq3btsmart.const import EQ3BT_MAX_OFFSET, EQ3BT_MIN_OFFSET
from eq3btsmart.exceptions import Eq3InvalidDataException


class Eq3TemperatureOffset(Adapter[int, int, float, float]):
    """Adapter to encode and decode temperature offset data."""

    def _encode(self, obj: float, ctx: Context, path: str) -> int:
        return self.encode(obj)

    def _decode(self, obj: int, ctx: Context, path: str) -> float:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: float) -> int:
        if value < EQ3BT_MIN_OFFSET or value > EQ3BT_MAX_OFFSET:
            raise Eq3InvalidDataException(
                f"Temperature {value} out of range [{EQ3BT_MIN_OFFSET}, {EQ3BT_MAX_OFFSET}]"
            )

        return int((value + 3.5) / 0.5)

    @classmethod
    def decode(cls, value: int) -> float:
        return float(value * 0.5 - 3.5)



from eq3btsmart.adapter.eq3_serial import Eq3Serial


def test_encode_valid_string() -> None:
    string = "example"
    encoded = Eq3Serial.encode(string)
    assert encoded == bytes([ord(char) + 0x30 for char in string])


def test_decode_valid_bytes() -> None:
    encoded_bytes = bytes([ord(char) + 0x30 for char in "example"])
    decoded = Eq3Serial.decode(encoded_bytes)
    assert decoded == "example"


def test_round_trip_consistency() -> None:
    string = "example"
    encoded = Eq3Serial.encode(string)
    decoded = Eq3Serial.decode(encoded)
    assert decoded == string


def test_encode_empty_string() -> None:
    string = ""
    encoded = Eq3Serial.encode(string)
    assert encoded == b""


def test_decode_empty_bytes() -> None:
    encoded_bytes = b""
    decoded = Eq3Serial.decode(encoded_bytes)
    assert decoded == ""



from datetime import datetime, timedelta

from construct_typed import Adapter, Context

from eq3btsmart.exceptions import Eq3InvalidDataException


class Eq3AwayTime(Adapter[bytes, bytes, datetime, datetime]):
    """Adapter to encode and decode away time data."""

    def _encode(self, obj: datetime, ctx: Context, path: str) -> bytes:
        return self.encode(obj)

    def _decode(self, obj: bytes, ctx: Context, path: str) -> datetime:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: datetime) -> bytes:
        value += timedelta(minutes=15)
        value -= timedelta(minutes=value.minute % 30)

        if value.year < 2000 or value.year > 2099:
            raise Eq3InvalidDataException("Invalid year, possible [2000, 2099]")

        year = value.year - 2000
        hour = value.hour * 2
        if value.minute != 0:
            hour |= 0x01

        return bytes([value.day, year, hour, value.month])

    @classmethod
    def decode(cls, value: bytes) -> datetime:
        if value == bytes([0x00, 0x00, 0x00, 0x00]):
            return datetime(year=2000, month=1, day=1, hour=0, minute=0)

        (day, year, hour_min, month) = value

        year += 2000

        minute = 0
        if hour_min & 0x01:
            minute = 30
        hour = int(hour_min / 2)

        return datetime(year=year, month=month, day=day, hour=hour, minute=minute)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Self

from eq3btsmart.structures import Eq3Struct


@dataclass
class BaseModel[StructType: Eq3Struct](ABC):
    @classmethod
    @abstractmethod
    def from_struct(cls: type[Self], struct: StructType) -> Self:
        """Convert the structure to a model."""

    @classmethod
    @abstractmethod
    def struct_type(cls: type[Self]) -> type[StructType]:
        """Return the structure type associated with the model."""

    @classmethod
    def from_bytes(cls: type[Self], data: bytes) -> Self:
        """Convert the data to a model."""
        return cls.from_struct(cls.struct_type().from_bytes(data))



from construct_typed import Adapter, Context

from eq3btsmart.const import (
    EQ3BT_OFF_TEMP,
    EQ3BT_ON_TEMP,
)
from eq3btsmart.exceptions import Eq3InvalidDataException


class Eq3Temperature(Adapter[int, int, float, float]):
    """Adapter to encode and decode temperature data."""

    def _encode(self, obj: float, ctx: Context, path: str) -> int:
        return self.encode(obj)

    def _decode(self, obj: int, ctx: Context, path: str) -> float:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: float) -> int:
        if value < EQ3BT_OFF_TEMP or value > EQ3BT_ON_TEMP:
            raise Eq3InvalidDataException(
                f"Temperature {value} out of range [{EQ3BT_OFF_TEMP}, {EQ3BT_ON_TEMP}]"
            )

        return int(value * 2)

    @classmethod
    def decode(cls, value: int) -> float:
        return float(value / 2)



import asyncio
from datetime import datetime, time, timedelta
from unittest.mock import MagicMock, patch

import pytest
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.exc import BleakError
from construct_typed import DataclassStruct

from eq3btsmart.const import (
    EQ3BT_ON_TEMP,
    Command,
    Eq3Characteristic,
    Eq3Event,
    Eq3Preset,
    OperationMode,
    WeekDay,
)
from eq3btsmart.exceptions import (
    Eq3AlreadyAwaitingResponseException,
    Eq3CommandException,
    Eq3ConnectionException,
    Eq3InternalException,
    Eq3StateException,
    Eq3TimeoutException,
)
from eq3btsmart.models.device_data import DeviceData
from eq3btsmart.models.presets import Presets
from eq3btsmart.models.schedule import Schedule
from eq3btsmart.models.schedule_day import ScheduleDay
from eq3btsmart.models.schedule_hour import ScheduleHour
from eq3btsmart.models.status import Status
from eq3btsmart.structures import Eq3Message, Eq3Struct, IdGetCommand
from eq3btsmart.thermostat import Thermostat
from tests.mock_client import MockClient, mock_id, mock_status


@pytest.mark.asyncio
async def test_connect_disconnect(mock_thermostat: Thermostat) -> None:
    assert isinstance(mock_thermostat._conn, MockClient)
    assert not mock_thermostat._conn.is_connected

    await mock_thermostat.async_connect()

    assert mock_thermostat._conn.is_connected

    await mock_thermostat.async_disconnect()

    assert not mock_thermostat._conn.is_connected


@pytest.mark.asyncio
async def test_get_id(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    initial_device_data = mock_thermostat._last_device_data

    mock_id.version = 100

    await mock_thermostat.async_get_device_data()

    assert mock_thermostat._last_device_data is not None
    assert isinstance(mock_thermostat._last_device_data.device_serial, str)
    assert mock_thermostat._last_device_data.device_serial == mock_id.serial
    assert initial_device_data != mock_thermostat._last_device_data


@pytest.mark.asyncio
async def test_get_status(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    initial_status = mock_thermostat._last_status

    mock_status.valve = 0x0F

    status = await mock_thermostat.async_get_status()

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.valve == 0x0F
    assert initial_status != mock_thermostat._last_status
    assert status is not None
    assert status.valve == 0x0F
    assert initial_status != status


@pytest.mark.asyncio
async def test_get_schedule(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    assert mock_thermostat._last_schedule is not None
    assert len(mock_thermostat._last_schedule.schedule_days) == 7

    schedule = await mock_thermostat.async_get_schedule()

    assert mock_thermostat._last_schedule is not None
    assert len(mock_thermostat._last_schedule.schedule_days) == 7
    assert schedule is not None
    assert len(schedule.schedule_days) == 7


@pytest.mark.asyncio
async def test_configure_window_open(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    await mock_thermostat.async_configure_window_open(6.5, timedelta(minutes=25))

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.presets is not None
    assert mock_thermostat._last_status.presets.window_open_temperature == 6.5
    assert mock_thermostat._last_status.presets.window_open_time == timedelta(
        minutes=25
    )


@pytest.mark.asyncio
async def test_configure_presets(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )

    await mock_thermostat.async_configure_presets(
        26.5,
        16,
    )

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.presets is not None
    assert mock_thermostat._last_status.presets.comfort_temperature == 26.5

    assert mock_thermostat._last_status.presets.eco_temperature == 16.0


@pytest.mark.asyncio
async def test_configure_comfort_without_status(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = None

    with pytest.raises(Exception):
        await mock_thermostat.async_configure_comfort_temperature(26.5)


@pytest.mark.asyncio
async def test_configure_eco_without_status(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = None

    with pytest.raises(Exception):
        await mock_thermostat.async_configure_eco_temperature(16)


@pytest.mark.asyncio
async def test_configure_temperature_offset(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    await mock_thermostat.async_configure_temperature_offset(2.5)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.presets is not None
    assert mock_thermostat._last_status.presets.offset_temperature == 2.5


@pytest.mark.asyncio
async def test_set_mode(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )

    await mock_thermostat.async_set_mode(OperationMode.AUTO)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.operation_mode == OperationMode.AUTO


@pytest.mark.asyncio
async def test_set_mode_without_status(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = None

    with pytest.raises(Eq3StateException):
        await mock_thermostat.async_set_mode(OperationMode.MANUAL)


@pytest.mark.asyncio
async def test_set_mode_manual(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.AUTO,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )

    await mock_thermostat.async_set_mode(OperationMode.MANUAL)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.operation_mode == OperationMode.MANUAL
    assert mock_thermostat._last_status.target_temperature == 21.0


@pytest.mark.asyncio
async def test_set_mode_off(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.AUTO,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )

    await mock_thermostat.async_set_mode(OperationMode.OFF)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.operation_mode == OperationMode.OFF
    assert mock_thermostat._last_status.target_temperature == 4.5


@pytest.mark.asyncio
async def test_set_mode_on(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.AUTO,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )

    await mock_thermostat.async_set_mode(OperationMode.ON)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.operation_mode == OperationMode.ON
    assert mock_thermostat._last_status.target_temperature == 30


@pytest.mark.asyncio
async def test_set_away(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    datetime_now = datetime.now()
    await mock_thermostat.async_set_away(datetime_now + timedelta(days=30), 21.5)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.is_away
    assert mock_thermostat._last_status.away_until is not None

    minute_difference = (
        mock_thermostat._last_status.away_until - (datetime_now + timedelta(days=30))
    ).total_seconds() / 60

    assert minute_difference <= 30


@pytest.mark.asyncio
async def test_set_away_not_disabled(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    await mock_thermostat.async_get_status()

    await mock_thermostat.async_set_mode(OperationMode.MANUAL)
    await mock_thermostat.async_set_away(datetime.now() + timedelta(days=30), 21.5)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.is_away
    assert mock_thermostat._last_status.operation_mode == OperationMode.AUTO


@pytest.mark.asyncio
async def test_set_temperature(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    await mock_thermostat.async_get_status()

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.target_temperature != 23.5

    await mock_thermostat.async_set_temperature(23.5)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.target_temperature == 23.5


@pytest.mark.asyncio
async def test_set_temperature_off(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    await mock_thermostat.async_get_status()

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.target_temperature != 4.5
    assert mock_thermostat._last_status.operation_mode != OperationMode.OFF

    await mock_thermostat.async_set_temperature(4.5)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.target_temperature == 4.5
    assert mock_thermostat._last_status.operation_mode.value == OperationMode.OFF.value


@pytest.mark.asyncio
async def test_set_temperature_on(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    await mock_thermostat.async_get_status()

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.target_temperature != 30
    assert mock_thermostat._last_status.operation_mode != OperationMode.ON

    await mock_thermostat.async_set_temperature(EQ3BT_ON_TEMP)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.target_temperature == 30
    assert mock_thermostat._last_status.operation_mode.value == OperationMode.ON.value


@pytest.mark.asyncio
async def test_set_preset_comfort(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    await mock_thermostat.async_get_status()
    await mock_thermostat.async_set_temperature(26)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.target_temperature == 26.0

    await mock_thermostat.async_configure_presets(
        comfort_temperature=21, eco_temperature=17
    )
    await mock_thermostat.async_set_preset(Eq3Preset.COMFORT)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.target_temperature == 21.0


@pytest.mark.asyncio
async def test_set_preset_eco(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    await mock_thermostat.async_get_status()
    await mock_thermostat.async_set_temperature(26)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.target_temperature == 26.0

    await mock_thermostat.async_configure_presets(
        comfort_temperature=21, eco_temperature=17
    )
    await mock_thermostat.async_set_preset(Eq3Preset.ECO)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.target_temperature == 17.0


@pytest.mark.asyncio
async def test_set_boost(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    await mock_thermostat.async_get_status()

    assert mock_thermostat._last_status is not None
    assert not mock_thermostat._last_status.is_boost
    assert mock_thermostat._last_status.operation_mode == OperationMode.MANUAL

    await mock_thermostat.async_set_boost(True)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.is_boost
    assert mock_thermostat._last_status.operation_mode == OperationMode.MANUAL

    await mock_thermostat.async_set_boost(False)

    assert mock_thermostat._last_status is not None
    assert not mock_thermostat._last_status.is_boost
    assert mock_thermostat._last_status.operation_mode == OperationMode.MANUAL


@pytest.mark.asyncio
async def test_set_locked(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    await mock_thermostat.async_get_status()

    assert mock_thermostat._last_status is not None
    assert not mock_thermostat._last_status.is_locked

    await mock_thermostat.async_set_locked(True)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.is_locked

    await mock_thermostat.async_set_locked(False)

    assert mock_thermostat._last_status is not None
    assert not mock_thermostat._last_status.is_locked


@pytest.mark.asyncio
async def test_set_schedule(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    await mock_thermostat.async_get_status()

    assert mock_thermostat._last_schedule is not None

    assert len(mock_thermostat._last_schedule.schedule_days) == 7

    schedule = Schedule(
        schedule_days=[
            ScheduleDay(
                WeekDay.MONDAY,
                schedule_hours=[
                    ScheduleHour(
                        21,
                        time(1, 0, 0),
                    )
                ],
            )
        ]
    )

    await mock_thermostat.async_set_schedule(schedule)

    assert mock_thermostat._last_schedule is not None
    assert schedule.schedule_days[0] in mock_thermostat._last_schedule.schedule_days


@pytest.mark.asyncio
async def test_delete_schedule(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    await mock_thermostat.async_get_status()

    assert mock_thermostat._last_schedule is not None
    assert len(mock_thermostat._last_schedule.schedule_days) == 7
    assert len(mock_thermostat._last_schedule.schedule_days[0].schedule_hours) == 1

    await mock_thermostat.async_delete_schedule(WeekDay.MONDAY)

    assert len(mock_thermostat._last_schedule.schedule_days) == 7
    assert len(mock_thermostat._last_schedule.schedule_days[2].schedule_hours) == 0

    await mock_thermostat.async_delete_schedule()

    assert len(mock_thermostat._last_schedule.schedule_days) == 7
    for schedule_day in mock_thermostat._last_schedule.schedule_days:
        assert len(schedule_day.schedule_hours) == 0


@pytest.mark.asyncio
async def test_write_not_connected(mock_thermostat: Thermostat) -> None:
    with pytest.raises(Exception):
        await mock_thermostat.async_set_boost(True)


@pytest.mark.asyncio
async def test_fail_on_invalid_notification(mock_thermostat: Thermostat) -> None:
    mock_characteristic = MagicMock(spec=BleakGATTCharacteristic)

    with pytest.raises(Exception):
        await mock_thermostat._on_message_received(
            mock_characteristic,
            bytearray(Eq3Message(cmd=Command.ID_RETURN, data=b"\x01").to_bytes()),
        )


@pytest.mark.asyncio
async def test_device_data(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    # Test when device data is set

    mock_id.version = 100
    await mock_thermostat.async_get_device_data()

    assert mock_thermostat.device_data is not None
    assert isinstance(mock_thermostat.device_data.device_serial, str)
    assert mock_thermostat.device_data.device_serial == mock_id.serial

    # Test when device data is not set
    mock_thermostat._last_device_data = None

    with pytest.raises(Eq3StateException, match="Device data not set"):
        _ = mock_thermostat.device_data


@pytest.mark.asyncio
async def test_presets(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    # Test when presets are set
    mock_thermostat._last_status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=Presets(
            comfort_temperature=21.0,
            eco_temperature=17.0,
            window_open_temperature=12.0,
            window_open_time=timedelta(minutes=15),
            offset_temperature=0.5,
        ),
    )

    presets = mock_thermostat.presets

    assert presets is not None
    assert presets.comfort_temperature == 21.0
    assert presets.eco_temperature == 17.0
    assert presets.window_open_temperature == 12.0
    assert presets.window_open_time == timedelta(minutes=15)
    assert presets.offset_temperature == 0.5

    # Test when presets are not set
    mock_thermostat._last_status.presets = None

    with pytest.raises(Eq3StateException, match="Presets not set"):
        _ = mock_thermostat.presets


@pytest.mark.asyncio
async def test_schedule(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    # Test when schedule is set
    mock_thermostat._last_schedule = Schedule(
        schedule_days=[
            ScheduleDay(
                WeekDay.MONDAY,
                schedule_hours=[
                    ScheduleHour(
                        target_temperature=21.0,
                        next_change_at=time(6, 0, 0),
                    )
                ],
            )
        ]
    )

    schedule = mock_thermostat.schedule

    assert schedule is not None
    assert len(schedule.schedule_days) == 1
    assert schedule.schedule_days[0].week_day == WeekDay.MONDAY
    assert len(schedule.schedule_days[0].schedule_hours) == 1
    assert schedule.schedule_days[0].schedule_hours[0].target_temperature == 21.0
    assert schedule.schedule_days[0].schedule_hours[0].next_change_at == time(6, 0, 0)

    # Test when schedule is not set
    mock_thermostat._last_schedule = None

    with pytest.raises(Eq3StateException, match="Schedule not set"):
        _ = mock_thermostat.schedule


@pytest.mark.asyncio
async def test_async_connect_success(mock_thermostat: Thermostat) -> None:
    with (
        patch.object(
            mock_thermostat._conn, "connect", return_value=None
        ) as mock_connect,
        patch.object(
            mock_thermostat._conn, "start_notify", return_value=None
        ) as mock_start_notify,
        patch.object(
            mock_thermostat, "async_get_device_data", return_value=MagicMock()
        ) as mock_get_device_data,
        patch.object(
            mock_thermostat, "async_get_status", return_value=MagicMock()
        ) as mock_get_status,
        patch.object(
            mock_thermostat, "async_get_schedule", return_value=MagicMock()
        ) as mock_get_schedule,
        patch.object(
            mock_thermostat, "trigger_event", return_value=None
        ) as mock_trigger_event,
    ):
        await mock_thermostat.async_connect()

        mock_connect.assert_called_once()
        mock_start_notify.assert_called_once_with(
            Eq3Characteristic.NOTIFY, mock_thermostat._on_message_received
        )
        mock_get_device_data.assert_called_once()
        mock_get_status.assert_called_once()
        mock_get_schedule.assert_called_once()
        mock_trigger_event.assert_called_once_with(
            Eq3Event.CONNECTED,
            device_data=mock_thermostat.device_data,
            status=mock_thermostat.status,
            schedule=mock_thermostat.schedule,
        )


@pytest.mark.asyncio
async def test_async_connect_bleak_error(mock_thermostat: Thermostat) -> None:
    with (
        patch.object(mock_thermostat._conn, "connect", side_effect=BleakError),
        patch.object(mock_thermostat._conn, "start_notify", return_value=None),
        patch.object(
            mock_thermostat, "async_get_device_data", return_value=MagicMock()
        ),
        patch.object(mock_thermostat, "async_get_status", return_value=MagicMock()),
        patch.object(mock_thermostat, "async_get_schedule", return_value=MagicMock()),
        patch.object(mock_thermostat, "trigger_event", return_value=None),
    ):
        with pytest.raises(
            Eq3ConnectionException, match="Could not connect to the device"
        ):
            await mock_thermostat.async_connect()


@pytest.mark.asyncio
async def test_async_connect_timeout_error(mock_thermostat: Thermostat) -> None:
    with (
        patch.object(mock_thermostat._conn, "connect", side_effect=TimeoutError),
        patch.object(mock_thermostat._conn, "start_notify", return_value=None),
        patch.object(
            mock_thermostat, "async_get_device_data", return_value=MagicMock()
        ),
        patch.object(mock_thermostat, "async_get_status", return_value=MagicMock()),
        patch.object(mock_thermostat, "async_get_schedule", return_value=MagicMock()),
        patch.object(mock_thermostat, "trigger_event", return_value=None),
    ):
        with pytest.raises(Eq3TimeoutException, match="Timeout during connection"):
            await mock_thermostat.async_connect()


@pytest.mark.asyncio
async def test_async_disconnect_success(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    with patch.object(
        mock_thermostat._conn, "disconnect", return_value=None
    ) as mock_disconnect:
        await mock_thermostat.async_disconnect()

        mock_disconnect.assert_called_once()


@pytest.mark.asyncio
async def test_async_disconnect_bleak_error(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    with patch.object(mock_thermostat._conn, "disconnect", side_effect=BleakError):
        with pytest.raises(
            Eq3ConnectionException, match="Could not disconnect from the device"
        ):
            await mock_thermostat.async_disconnect()


@pytest.mark.asyncio
async def test_async_disconnect_timeout_error(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    with patch.object(mock_thermostat._conn, "disconnect", side_effect=TimeoutError):
        with pytest.raises(Eq3TimeoutException, match="Timeout during disconnection"):
            await mock_thermostat.async_disconnect()


@pytest.mark.asyncio
async def test_async_disconnect_with_pending_futures(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._device_data_future = asyncio.Future()
    mock_thermostat._status_future = asyncio.Future()
    mock_thermostat._schedule_future = asyncio.Future()

    with patch.object(mock_thermostat._conn, "disconnect", return_value=None):
        await mock_thermostat.async_disconnect()

        assert mock_thermostat._device_data_future.exception() is not None
        assert isinstance(
            mock_thermostat._device_data_future.exception(), Eq3ConnectionException
        )
        assert mock_thermostat._status_future.exception() is not None
        assert isinstance(
            mock_thermostat._status_future.exception(), Eq3ConnectionException
        )
        assert mock_thermostat._schedule_future.exception() is not None
        assert isinstance(
            mock_thermostat._schedule_future.exception(), Eq3ConnectionException
        )


@pytest.mark.asyncio
async def test_is_connected(mock_thermostat: Thermostat) -> None:
    # Initially, the thermostat should not be connected
    assert not mock_thermostat.is_connected

    # Connect the thermostat
    await mock_thermostat.async_connect()

    # Now, the thermostat should be connected
    assert mock_thermostat.is_connected

    # Disconnect the thermostat
    await mock_thermostat.async_disconnect()

    # Finally, the thermostat should not be connected again
    assert not mock_thermostat.is_connected


@pytest.mark.asyncio
async def test_configure_window_open_temperature(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=Presets(
            comfort_temperature=21.0,
            eco_temperature=17.0,
            window_open_temperature=12.0,
            window_open_time=timedelta(minutes=15),
            offset_temperature=0.5,
        ),
    )

    await mock_thermostat.async_configure_window_open_temperature(6.5)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.presets is not None
    assert mock_thermostat._last_status.presets.window_open_temperature == 6.5
    assert mock_thermostat._last_status.presets.window_open_time == timedelta(
        minutes=15
    )


@pytest.mark.asyncio
async def test_configure_window_open_temperature_without_presets(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )

    with pytest.raises(Eq3StateException, match="Presets not set"):
        await mock_thermostat.async_configure_window_open_temperature(6.5)


@pytest.mark.asyncio
async def test_configure_window_open_duration(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=Presets(
            comfort_temperature=21.0,
            eco_temperature=17.0,
            window_open_temperature=12.0,
            window_open_time=timedelta(minutes=15),
            offset_temperature=0.5,
        ),
    )

    await mock_thermostat.async_configure_window_open_duration(timedelta(minutes=30))

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.presets is not None
    assert mock_thermostat._last_status.presets.window_open_time == timedelta(
        minutes=30
    )


@pytest.mark.asyncio
async def test_configure_window_open_duration_float(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=Presets(
            comfort_temperature=21.0,
            eco_temperature=17.0,
            window_open_temperature=12.0,
            window_open_time=timedelta(minutes=15),
            offset_temperature=0.5,
        ),
    )

    await mock_thermostat.async_configure_window_open_duration(45.0)

    assert mock_thermostat._last_status is not None
    assert mock_thermostat._last_status.presets is not None
    assert mock_thermostat._last_status.presets.window_open_time == timedelta(
        minutes=45
    )


@pytest.mark.asyncio
async def test_configure_window_open_duration_without_presets(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    mock_thermostat._last_status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )

    with pytest.raises(Eq3StateException, match="Presets not set"):
        await mock_thermostat.async_configure_window_open_duration(
            timedelta(minutes=30)
        )


@pytest.mark.asyncio
async def test_aenter(mock_thermostat: Thermostat) -> None:
    async with mock_thermostat:
        assert mock_thermostat.is_connected


@pytest.mark.asyncio
async def test_aexit(mock_thermostat: Thermostat) -> None:
    async with mock_thermostat:
        pass
    assert not mock_thermostat.is_connected


@pytest.mark.asyncio
async def test_async_write_command_with_device_data_response_success(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    command = IdGetCommand()

    result = await mock_thermostat._async_write_command_with_device_data_response(
        command
    )
    assert result == DeviceData.from_struct(mock_id)


@pytest.mark.asyncio
async def test_async_write_command_with_device_data_response_already_awaiting(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    command = MagicMock(spec=Eq3Struct)
    mock_thermostat._device_data_future = asyncio.Future()

    with pytest.raises(
        Eq3AlreadyAwaitingResponseException,
        match="Already awaiting a device data command response",
    ):
        await mock_thermostat._async_write_command_with_device_data_response(command)


@pytest.mark.asyncio
async def test_async_write_command_with_device_data_response_timeout(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    command = IdGetCommand()

    with (
        patch.object(mock_thermostat, "_async_write_command", return_value=None),
        patch("asyncio.wait_for", side_effect=TimeoutError),
    ):
        with pytest.raises(
            Eq3TimeoutException, match="Timeout during device data command"
        ):
            await mock_thermostat._async_write_command_with_device_data_response(
                command
            )


@pytest.mark.asyncio
async def test_async_write_command_with_status_response_success(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    command = MagicMock(spec=Eq3Struct)
    mock_status = MagicMock(spec=Status)

    with (
        patch.object(mock_thermostat, "_async_write_command", return_value=None),
        patch("asyncio.wait_for", return_value=mock_status),
    ):
        result = await mock_thermostat._async_write_command_with_status_response(
            command
        )
        assert result == mock_status


@pytest.mark.asyncio
async def test_async_write_command_with_status_response_already_awaiting(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    command = MagicMock(spec=Eq3Struct)
    mock_thermostat._status_future = asyncio.Future()

    with pytest.raises(
        Eq3AlreadyAwaitingResponseException,
        match="Already awaiting a status command response",
    ):
        await mock_thermostat._async_write_command_with_status_response(command)


@pytest.mark.asyncio
async def test_async_write_command_with_status_response_timeout(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    command = MagicMock(spec=Eq3Struct)

    with (
        patch.object(mock_thermostat, "_async_write_command", return_value=None),
        patch("asyncio.wait_for", side_effect=TimeoutError),
    ):
        with pytest.raises(Eq3TimeoutException, match="Timeout during status command"):
            await mock_thermostat._async_write_command_with_status_response(command)


@pytest.mark.asyncio
async def test_async_write_commands_with_schedule_response_success(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    commands = [MagicMock(spec=Eq3Struct) for _ in range(3)]
    mock_schedule = MagicMock(spec=Schedule)

    with (
        patch.object(mock_thermostat, "_async_write_command", return_value=None),
        patch("asyncio.wait_for", return_value=mock_schedule),
    ):
        result = await mock_thermostat._async_write_commands_with_schedule_response(
            commands  # type: ignore
        )
        assert result == mock_schedule


@pytest.mark.asyncio
async def test_async_write_commands_with_schedule_response_already_awaiting(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    commands = [MagicMock(spec=Eq3Struct) for _ in range(3)]
    mock_thermostat._schedule_future = asyncio.Future()

    with pytest.raises(
        Eq3AlreadyAwaitingResponseException,
        match="Already awaiting a schedule command response",
    ):
        await mock_thermostat._async_write_commands_with_schedule_response(commands)  # type: ignore


@pytest.mark.asyncio
async def test_async_write_commands_with_schedule_response_timeout(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    commands = [MagicMock(spec=Eq3Struct) for _ in range(3)]

    with (
        patch.object(mock_thermostat, "_async_write_command", return_value=None),
        patch("asyncio.wait_for", side_effect=TimeoutError),
    ):
        with pytest.raises(
            Eq3TimeoutException, match="Timeout during schedule command"
        ):
            await mock_thermostat._async_write_commands_with_schedule_response(commands)  # type: ignore


@pytest.mark.asyncio
async def test_async_write_command_success(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    command = MagicMock(spec=Eq3Struct)
    command.to_bytes.return_value = b"\x01\x02\x03"

    with patch.object(
        mock_thermostat._conn, "write_gatt_char", return_value=None
    ) as mock_write_gatt_char:
        await mock_thermostat._async_write_command(command)

        mock_write_gatt_char.assert_called_once_with(
            Eq3Characteristic.WRITE, b"\x01\x02\x03"
        )


@pytest.mark.asyncio
async def test_async_write_command_not_connected(mock_thermostat: Thermostat) -> None:
    command = MagicMock(spec=Eq3Struct)

    with pytest.raises(Eq3ConnectionException, match="Not connected"):
        await mock_thermostat._async_write_command(command)


@pytest.mark.asyncio
async def test_async_write_command_bleak_error(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    command = MagicMock(spec=Eq3Struct)
    command.to_bytes.return_value = b"\x01\x02\x03"

    with patch.object(mock_thermostat._conn, "write_gatt_char", side_effect=BleakError):
        with pytest.raises(Eq3CommandException, match="Error during write"):
            await mock_thermostat._async_write_command(command)


@pytest.mark.asyncio
async def test_async_write_command_timeout_error(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    command = MagicMock(spec=Eq3Struct)
    command.to_bytes.return_value = b"\x01\x02\x03"

    with patch.object(
        mock_thermostat._conn, "write_gatt_char", side_effect=TimeoutError
    ):
        with pytest.raises(Eq3TimeoutException, match="Timeout during write"):
            await mock_thermostat._async_write_command(command)


@pytest.mark.asyncio
async def test_on_message_received_id_return(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_characteristic = MagicMock(spec=BleakGATTCharacteristic)
    data = Eq3Message(
        cmd=Command.ID_RETURN, is_status_command=True, data=b"\x02\x03"
    ).to_bytes()

    with (
        patch.object(
            mock_thermostat, "_on_device_data_received", return_value=None
        ) as mock_on_device_data_received,
        patch.object(
            DeviceData, "from_bytes", return_value=DeviceData.from_struct(mock_id)
        ),
        patch.object(
            DeviceData, "from_struct", return_value=DeviceData.from_struct(mock_id)
        ),
    ):
        await mock_thermostat._on_message_received(mock_characteristic, bytearray(data))

        mock_on_device_data_received.assert_called_once()
        assert isinstance(mock_on_device_data_received.call_args[0][0], DeviceData)


@pytest.mark.asyncio
async def test_on_message_received_info_return(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_characteristic = MagicMock(spec=BleakGATTCharacteristic)
    data = Eq3Message(
        cmd=Command.INFO_RETURN, is_status_command=True, data=b"\x02\x03"
    ).to_bytes()

    with (
        patch.object(
            mock_thermostat, "_on_status_received", return_value=None
        ) as mock_on_status_received,
        patch.object(
            Status, "from_bytes", return_value=Status.from_struct(mock_status)
        ),
        patch.object(
            Status, "from_struct", return_value=Status.from_struct(mock_status)
        ),
    ):
        await mock_thermostat._on_message_received(mock_characteristic, bytearray(data))

        mock_on_status_received.assert_called_once()
        assert isinstance(mock_on_status_received.call_args[0][0], Status)


@pytest.mark.asyncio
async def test_on_message_received_schedule_return(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_characteristic = MagicMock(spec=BleakGATTCharacteristic)
    data = Eq3Message(
        cmd=Command.SCHEDULE_RETURN, is_status_command=True, data=b"\x02\x03"
    ).to_bytes()

    with (
        patch.object(
            mock_thermostat, "_on_schedule_received", return_value=None
        ) as mock_on_schedule_received,
        patch.object(
            Schedule,
            "from_bytes",
            return_value=Schedule(),
        ),
    ):
        await mock_thermostat._on_message_received(mock_characteristic, bytearray(data))

        mock_on_schedule_received.assert_called_once()
        assert isinstance(mock_on_schedule_received.call_args[0][0], Schedule)


@pytest.mark.asyncio
async def test_on_message_received_unknown_command(mock_thermostat: Thermostat) -> None:
    await mock_thermostat.async_connect()

    mock_characteristic = MagicMock(spec=BleakGATTCharacteristic)
    data = MagicMock(spec=bytearray)
    command = MagicMock(spec=Eq3Message)
    command.cmd = 0xCC
    command.is_status_command = True
    command.data = data

    with (
        patch.object(DataclassStruct, "parse", return_value=command),
    ):
        with pytest.raises(Eq3InternalException, match="Unknown command"):
            await mock_thermostat._on_message_received(mock_characteristic, data)


@pytest.mark.asyncio
async def test_on_message_received_info_return_not_status_command(
    mock_thermostat: Thermostat,
) -> None:
    await mock_thermostat.async_connect()

    mock_characteristic = MagicMock(spec=BleakGATTCharacteristic)
    data = Eq3Message(
        cmd=Command.INFO_RETURN, is_status_command=False, data=b"\x02\x03"
    ).to_bytes()

    with (
        patch.object(
            mock_thermostat, "_on_status_received", return_value=None
        ) as mock_on_status_received,
        patch.object(
            Status, "from_bytes", return_value=Status.from_struct(mock_status)
        ),
        patch.object(
            Status, "from_struct", return_value=Status.from_struct(mock_status)
        ),
    ):
        await mock_thermostat._on_message_received(mock_characteristic, bytearray(data))

        mock_on_status_received.assert_not_called()


@pytest.mark.asyncio
async def test_trigger_event_disconnected(mock_thermostat: Thermostat) -> None:
    callback = MagicMock()
    mock_thermostat.register_callback(Eq3Event.DISCONNECTED, callback)

    await mock_thermostat.trigger_event(Eq3Event.DISCONNECTED)

    callback.assert_called_once_with()


@pytest.mark.asyncio
async def test_trigger_event_connected(mock_thermostat: Thermostat) -> None:
    callback = MagicMock()
    mock_thermostat.register_callback(Eq3Event.CONNECTED, callback)

    device_data = DeviceData.from_struct(mock_id)
    status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )
    schedule = Schedule(
        schedule_days=[
            ScheduleDay(
                WeekDay.MONDAY,
                schedule_hours=[
                    ScheduleHour(
                        target_temperature=21.0,
                        next_change_at=time(6, 0, 0),
                    )
                ],
            )
        ]
    )

    await mock_thermostat.trigger_event(
        Eq3Event.CONNECTED, device_data=device_data, status=status, schedule=schedule
    )

    callback.assert_called_once_with(device_data, status, schedule)


@pytest.mark.asyncio
async def test_trigger_event_device_data_received(mock_thermostat: Thermostat) -> None:
    callback = MagicMock()
    mock_thermostat.register_callback(Eq3Event.DEVICE_DATA_RECEIVED, callback)

    device_data = DeviceData.from_struct(mock_id)

    await mock_thermostat.trigger_event(
        Eq3Event.DEVICE_DATA_RECEIVED, device_data=device_data
    )

    callback.assert_called_once_with(device_data)


@pytest.mark.asyncio
async def test_trigger_event_status_received(mock_thermostat: Thermostat) -> None:
    callback = MagicMock()
    mock_thermostat.register_callback(Eq3Event.STATUS_RECEIVED, callback)

    status = Status(
        valve=10,
        target_temperature=21.0,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )

    await mock_thermostat.trigger_event(Eq3Event.STATUS_RECEIVED, status=status)

    callback.assert_called_once_with(status)


@pytest.mark.asyncio
async def test_trigger_event_schedule_received(mock_thermostat: Thermostat) -> None:
    callback = MagicMock()
    mock_thermostat.register_callback(Eq3Event.SCHEDULE_RECEIVED, callback)

    schedule = Schedule(
        schedule_days=[
            ScheduleDay(
                WeekDay.MONDAY,
                schedule_hours=[
                    ScheduleHour(
                        target_temperature=21.0,
                        next_change_at=time(6, 0, 0),
                    )
                ],
            )
        ]
    )

    await mock_thermostat.trigger_event(Eq3Event.SCHEDULE_RECEIVED, schedule=schedule)

    callback.assert_called_once_with(schedule)


@pytest.mark.asyncio
async def test_on_device_data_received_no_future(mock_thermostat: Thermostat) -> None:
    mock_device_data = MagicMock(spec=DeviceData)
    mock_thermostat._device_data_future = None

    with patch.object(
        mock_thermostat, "trigger_event", return_value=None
    ) as mock_trigger_event:
        await mock_thermostat._on_device_data_received(mock_device_data)

        mock_trigger_event.assert_called_once_with(
            Eq3Event.DEVICE_DATA_RECEIVED, device_data=mock_device_data
        )


@pytest.mark.asyncio
async def test_on_status_received_no_future(mock_thermostat: Thermostat) -> None:
    mock_status = MagicMock(spec=Status)
    mock_thermostat._status_future = None

    with patch.object(
        mock_thermostat, "trigger_event", return_value=None
    ) as mock_trigger_event:
        await mock_thermostat._on_status_received(mock_status)

        mock_trigger_event.assert_called_once_with(
            Eq3Event.STATUS_RECEIVED, status=mock_status
        )


@pytest.mark.asyncio
async def test_on_schedule_received_no_future(mock_thermostat: Thermostat) -> None:
    mock_schedule = MagicMock(spec=Schedule)
    mock_thermostat._schedule_future = None
    mock_thermostat._last_schedule = None

    with patch.object(
        mock_thermostat, "trigger_event", return_value=None
    ) as mock_trigger_event:
        await mock_thermostat._on_schedule_received(mock_schedule)

        mock_trigger_event.assert_called_once_with(
            Eq3Event.SCHEDULE_RECEIVED, schedule=mock_schedule
        )


@pytest.mark.asyncio
async def test_register_callback(mock_thermostat: Thermostat) -> None:
    async def async_callback() -> None: ...

    def sync_callback() -> None: ...

    # Register async callback
    mock_thermostat.register_callback(Eq3Event.DISCONNECTED, async_callback)
    assert async_callback in mock_thermostat._callbacks[Eq3Event.DISCONNECTED]

    # Register sync callback
    mock_thermostat.register_callback(Eq3Event.DISCONNECTED, sync_callback)
    assert sync_callback in mock_thermostat._callbacks[Eq3Event.DISCONNECTED]

    # Register the same callback again, should not duplicate
    mock_thermostat.register_callback(Eq3Event.DISCONNECTED, sync_callback)
    assert mock_thermostat._callbacks[Eq3Event.DISCONNECTED].count(sync_callback) == 1


@pytest.mark.asyncio
async def test_unregister_callback(mock_thermostat: Thermostat) -> None:
    async def async_callback() -> None: ...

    def sync_callback() -> None: ...

    # Register callbacks
    mock_thermostat.register_callback(Eq3Event.DISCONNECTED, async_callback)
    mock_thermostat.register_callback(Eq3Event.DISCONNECTED, sync_callback)

    # Unregister async callback
    mock_thermostat.unregister_callback(Eq3Event.DISCONNECTED, async_callback)
    assert async_callback not in mock_thermostat._callbacks[Eq3Event.DISCONNECTED]

    # Unregister sync callback
    mock_thermostat.unregister_callback(Eq3Event.DISCONNECTED, sync_callback)
    assert sync_callback not in mock_thermostat._callbacks[Eq3Event.DISCONNECTED]

    # Unregister a callback that is not registered, should not raise an error
    mock_thermostat.unregister_callback(Eq3Event.DISCONNECTED, sync_callback)


@pytest.mark.asyncio
async def test_trigger_event_connected_invalid_arguments(
    mock_thermostat: Thermostat,
) -> None:
    with pytest.raises(
        Eq3InternalException,
        match="device_data, status, and schedule must not be None for CONNECTED event",
    ):
        await mock_thermostat.trigger_event(Eq3Event.CONNECTED)  # type: ignore


@pytest.mark.asyncio
async def test_trigger_event_device_data_received_invalid_arguments(
    mock_thermostat: Thermostat,
) -> None:
    with pytest.raises(
        Eq3InternalException,
        match="device_data must not be None for DEVICE_DATA_RECEIVED event",
    ):
        await mock_thermostat.trigger_event(Eq3Event.DEVICE_DATA_RECEIVED)  # type: ignore


@pytest.mark.asyncio
async def test_trigger_event_status_received_invalid_arguments(
    mock_thermostat: Thermostat,
) -> None:
    with pytest.raises(
        Eq3InternalException, match="status must not be None for STATUS_RECEIVED event"
    ):
        await mock_thermostat.trigger_event(Eq3Event.STATUS_RECEIVED)  # type: ignore


@pytest.mark.asyncio
async def test_trigger_event_schedule_received_invalid_arguments(
    mock_thermostat: Thermostat,
) -> None:
    with pytest.raises(
        Eq3InternalException,
        match="schedule must not be None for SCHEDULE_RECEIVED event",
    ):
        await mock_thermostat.trigger_event(Eq3Event.SCHEDULE_RECEIVED)  # type: ignore



from datetime import datetime, timedelta

import pytest

from eq3btsmart.adapter.eq3_away_time import Eq3AwayTime
from eq3btsmart.exceptions import Eq3InvalidDataException


def test_encode_valid_datetime() -> None:
    dt = datetime(year=2022, month=6, day=15, hour=10, minute=45)
    encoded = Eq3AwayTime.encode(dt)
    expected_bytes = bytes([15, 22, 22, 6])
    assert encoded == expected_bytes


def test_encode_valid_datetime_with_minute() -> None:
    dt = datetime(year=2022, month=6, day=15, hour=10, minute=25)
    encoded = Eq3AwayTime.encode(dt)
    expected_bytes = bytes([15, 22, 21, 6])
    assert encoded == expected_bytes


def test_decode_valid_bytes() -> None:
    encoded_bytes = bytes([15, 22, 21, 6])
    decoded = Eq3AwayTime.decode(encoded_bytes)
    expected_datetime = datetime(year=2022, month=6, day=15, hour=10, minute=30)
    assert decoded == expected_datetime


def test_round_trip_consistency() -> None:
    dt = datetime(year=2022, month=6, day=15, hour=10, minute=45)
    encoded = Eq3AwayTime.encode(dt)
    decoded = Eq3AwayTime.decode(encoded)
    dt_adjusted = dt + timedelta(minutes=15)
    dt_adjusted -= timedelta(minutes=dt_adjusted.minute % 30)
    assert decoded == dt_adjusted


def test_decode_special_case() -> None:
    encoded_bytes = bytes([0x00, 0x00, 0x00, 0x00])
    decoded = Eq3AwayTime.decode(encoded_bytes)
    expected_datetime = datetime(year=2000, month=1, day=1, hour=0, minute=0)
    assert decoded == expected_datetime


def test_encode_invalid_year() -> None:
    dt = datetime(year=2100, month=6, day=15, hour=10, minute=45)
    with pytest.raises(Eq3InvalidDataException):
        Eq3AwayTime.encode(dt)



from datetime import timedelta

import pytest

from eq3btsmart.adapter.eq3_duration import Eq3Duration
from eq3btsmart.exceptions import Eq3InvalidDataException


def test_encode_valid_timedelta() -> None:
    td = timedelta(minutes=25)
    encoded = Eq3Duration.encode(td)
    assert encoded == 5


def test_decode_valid_int() -> None:
    encoded_int = 5
    decoded = Eq3Duration.decode(encoded_int)
    assert decoded == timedelta(minutes=25)


def test_round_trip_consistency() -> None:
    td = timedelta(minutes=25)
    encoded = Eq3Duration.encode(td)
    decoded = Eq3Duration.decode(encoded)
    assert decoded == td


def test_encode_invalid_timedelta_negative() -> None:
    td = timedelta(minutes=-5)
    with pytest.raises(Eq3InvalidDataException):
        Eq3Duration.encode(td)


def test_encode_invalid_timedelta_exceed() -> None:
    td = timedelta(minutes=65)
    with pytest.raises(Eq3InvalidDataException):
        Eq3Duration.encode(td)



from datetime import datetime

import pytest

from eq3btsmart.adapter.eq3_time import Eq3Time
from eq3btsmart.exceptions import Eq3InvalidDataException


def test_encode_valid_datetime() -> None:
    dt = datetime(year=2022, month=6, day=15, hour=10, minute=45, second=30)
    encoded = Eq3Time.encode(dt)
    assert encoded == bytes([22, 6, 15, 10, 45, 30])


def test_decode_valid_bytes() -> None:
    encoded_bytes = bytes([22, 6, 15, 10, 45, 30])
    decoded = Eq3Time.decode(encoded_bytes)
    assert decoded == datetime(
        year=2022, month=6, day=15, hour=10, minute=45, second=30
    )


def test_round_trip_consistency() -> None:
    dt = datetime(year=2022, month=6, day=15, hour=10, minute=45, second=30)
    encoded = Eq3Time.encode(dt)
    decoded = Eq3Time.decode(encoded)
    assert decoded == dt


def test_decode_invalid_bytes_length() -> None:
    encoded_bytes = bytes([22, 6, 15, 10, 45])
    with pytest.raises(Eq3InvalidDataException):
        Eq3Time.decode(encoded_bytes)


def test_encode_future_date() -> None:
    dt = datetime(year=2122, month=6, day=15, hour=10, minute=45, second=30)
    encoded = Eq3Time.encode(dt)
    assert encoded == bytes([22, 6, 15, 10, 45, 30])



"""Exceptions for the eq3btsmart library."""


class Eq3Exception(Exception):
    """Base exception for the eq3btsmart library."""


class Eq3ConnectionException(Eq3Exception):
    """Exception for connection errors."""


class Eq3CommandException(Eq3Exception):
    """Exception for command errors."""


class Eq3AlreadyAwaitingResponseException(Eq3Exception):
    """Exception for commands that are already awaiting a response."""


class Eq3TimeoutException(Eq3Exception):
    """Exception for timeouts."""


class Eq3StateException(Eq3Exception):
    """Exception for invalid states."""


class Eq3InternalException(Eq3Exception):
    """Exception for internal errors."""


class Eq3InvalidDataException(Eq3Exception):
    """Exception for invalid data."""



from eq3btsmart.const import WeekDay


def test_from_index() -> None:
    assert WeekDay.from_index(0) == WeekDay.MONDAY
    assert WeekDay.from_index(1) == WeekDay.TUESDAY
    assert WeekDay.from_index(2) == WeekDay.WEDNESDAY
    assert WeekDay.from_index(3) == WeekDay.THURSDAY
    assert WeekDay.from_index(4) == WeekDay.FRIDAY
    assert WeekDay.from_index(5) == WeekDay.SATURDAY
    assert WeekDay.from_index(6) == WeekDay.SUNDAY



from datetime import time

from eq3btsmart.adapter.eq3_schedule_time import Eq3ScheduleTime


def test_encode_valid_time() -> None:
    t = time(hour=13, minute=30)
    encoded = Eq3ScheduleTime.encode(t)
    assert encoded == 81


def test_decode_valid_int() -> None:
    encoded_int = 81
    decoded = Eq3ScheduleTime.decode(encoded_int)
    assert decoded == time(hour=13, minute=30)


def test_round_trip_consistency() -> None:
    t = time(hour=13, minute=30)
    encoded = Eq3ScheduleTime.encode(t)
    decoded = Eq3ScheduleTime.decode(encoded)
    assert decoded == t


def test_encode_edge_case() -> None:
    t = time(hour=23, minute=59)
    encoded = Eq3ScheduleTime.encode(t)
    assert encoded == 143


def test_decode_edge_case() -> None:
    encoded_int = 143
    decoded = Eq3ScheduleTime.decode(encoded_int)
    assert decoded == time(hour=23, minute=50)



"""Support for eq3 Bluetooth Smart thermostats."""

import asyncio
from collections import defaultdict
from datetime import datetime, timedelta
from types import TracebackType
from typing import Awaitable, Callable, Literal, Self, Union, overload

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.exc import BleakError
from construct_typed import DataclassStruct

from eq3btsmart.adapter.eq3_temperature import Eq3Temperature
from eq3btsmart.const import (
    DEFAULT_COMMAND_TIMEOUT,
    DEFAULT_CONNECTION_TIMEOUT,
    EQ3BT_OFF_TEMP,
    EQ3BT_ON_TEMP,
    Command,
    Eq3Characteristic,
    Eq3Event,
    Eq3Preset,
    OperationMode,
    WeekDay,
)
from eq3btsmart.exceptions import (
    Eq3AlreadyAwaitingResponseException,
    Eq3CommandException,
    Eq3ConnectionException,
    Eq3InternalException,
    Eq3StateException,
    Eq3TimeoutException,
)
from eq3btsmart.models import DeviceData, Schedule, Status
from eq3btsmart.models.presets import Presets
from eq3btsmart.structures import (
    AwaySetCommand,
    BoostSetCommand,
    ComfortEcoConfigureCommand,
    ComfortSetCommand,
    EcoSetCommand,
    Eq3Message,
    Eq3Struct,
    IdGetCommand,
    InfoGetCommand,
    LockSetCommand,
    ModeSetCommand,
    OffsetConfigureCommand,
    ScheduleGetCommand,
    ScheduleHourStruct,
    ScheduleSetCommand,
    TemperatureSetCommand,
    WindowOpenConfigureCommand,
)


class Thermostat:
    """Representation of a EQ3 Bluetooth Smart thermostat."""

    def __init__(
        self,
        mac_address: str,
        connection_timeout: int = DEFAULT_CONNECTION_TIMEOUT,
        command_timeout: int = DEFAULT_COMMAND_TIMEOUT,
    ):
        """Initialize the thermostat."""

        self.mac_address = mac_address

        self._last_status: Status | None = None
        self._last_device_data: DeviceData | None = None
        self._last_schedule: Schedule | None = None

        self._callbacks: defaultdict[
            Eq3Event, list[Union[Callable[..., None], Callable[..., Awaitable[None]]]]
        ] = defaultdict(list)
        self._conn: BleakClient = BleakClient(
            mac_address,
            disconnected_callback=self._on_disconnected,
            timeout=DEFAULT_CONNECTION_TIMEOUT,
        )
        self._device_data_future: asyncio.Future[DeviceData] | None = None
        self._status_future: asyncio.Future[Status] | None = None
        self._schedule_future: asyncio.Future[Schedule] | None = None
        self._schedule_future_counter: int = 0
        self._lock = asyncio.Lock()
        self._connection_timeout = connection_timeout
        self._command_timeout = command_timeout

    @property
    def is_connected(self) -> bool:
        """Check if the thermostat is connected."""
        return self._conn.is_connected

    @property
    def device_data(self) -> DeviceData:
        """Get the last known device data, ensuring it's not None."""
        if self._last_device_data is None:
            raise Eq3StateException("Device data not set")
        return self._last_device_data

    @property
    def status(self) -> Status:
        """Get the last known status, ensuring it's not None."""
        if self._last_status is None:
            raise Eq3StateException("Status not set")
        return self._last_status

    @property
    def presets(self) -> Presets:
        """Get the presets from the last known status."""

        if self.status.presets is None:
            raise Eq3StateException("Presets not set")
        return self.status.presets

    @property
    def schedule(self) -> Schedule:
        """Get the last known schedule, ensuring it's not None."""
        if self._last_schedule is None:
            raise Eq3StateException("Schedule not set")
        return self._last_schedule

    async def async_connect(self) -> None:
        """Connect to the thermostat."""

        try:
            await asyncio.wait_for(self._conn.connect(), self._connection_timeout)
            await self._conn.start_notify(
                Eq3Characteristic.NOTIFY, self._on_message_received
            )
            (
                self._last_device_data,
                self._last_status,
                self._last_schedule,
            ) = await asyncio.gather(
                self.async_get_device_data(),
                self.async_get_status(),
                self.async_get_schedule(),
            )
        except BleakError as ex:
            raise Eq3ConnectionException("Could not connect to the device") from ex
        except TimeoutError as ex:
            raise Eq3TimeoutException("Timeout during connection") from ex

        await self.trigger_event(
            Eq3Event.CONNECTED,
            device_data=self.device_data,
            status=self.status,
            schedule=self.schedule,
        )

    async def async_disconnect(self) -> None:
        """Shutdown the connection to the thermostat."""

        exception = Eq3ConnectionException("Connection closed")
        if self._device_data_future is not None and not self._device_data_future.done():
            self._device_data_future.set_exception(exception)

        if self._status_future is not None and not self._status_future.done():
            self._status_future.set_exception(exception)

        if self._schedule_future is not None and not self._schedule_future.done():
            self._schedule_future.set_exception(exception)

        try:
            await self._conn.disconnect()
        except BleakError as ex:
            raise Eq3ConnectionException("Could not disconnect from the device") from ex
        except TimeoutError as ex:
            raise Eq3TimeoutException("Timeout during disconnection") from ex

    async def async_get_device_data(self) -> DeviceData:
        """Query device identification information, e.g. the serial number."""

        return await self._async_write_command_with_device_data_response(IdGetCommand())

    async def async_get_status(self) -> Status:
        """Query the thermostat status."""

        return await self._async_write_command_with_status_response(
            InfoGetCommand(time=datetime.now())
        )

    async def async_get_schedule(self) -> Schedule:
        """Query the schedule."""

        return await self._async_write_commands_with_schedule_response(
            [ScheduleGetCommand(day=week_day) for week_day in WeekDay]
        )

    async def async_configure_window_open(
        self, temperature: float, duration: timedelta
    ) -> Status:
        """Configures the window open behavior. The duration is specified in 5 minute increments."""

        return await self._async_write_command_with_status_response(
            WindowOpenConfigureCommand(
                window_open_temperature=temperature,
                window_open_time=duration,
            )
        )

    async def async_configure_window_open_temperature(
        self, temperature: float
    ) -> Status:
        """Configures the window open temperature."""

        return await self.async_configure_window_open(
            temperature, self.presets.window_open_time
        )

    async def async_configure_window_open_duration(
        self,
        duration: timedelta | float,
    ) -> Status:
        """Configures the window open duration."""

        if isinstance(duration, float):
            duration = timedelta(minutes=duration)

        return await self.async_configure_window_open(
            self.presets.window_open_temperature, duration
        )

    async def async_configure_presets(
        self,
        comfort_temperature: float,
        eco_temperature: float,
    ) -> Status:
        """Set the thermostats preset temperatures comfort (sun) and eco (moon)."""

        return await self._async_write_command_with_status_response(
            ComfortEcoConfigureCommand(
                comfort_temperature=comfort_temperature,
                eco_temperature=eco_temperature,
            )
        )

    async def async_configure_comfort_temperature(
        self, comfort_temperature: float
    ) -> Status:
        """Sets the thermostat's comfort temperature."""

        return await self.async_configure_presets(
            comfort_temperature, self.presets.eco_temperature
        )

    async def async_configure_eco_temperature(self, eco_temperature: float) -> Status:
        """Sets the thermostat's eco temperature."""

        return await self.async_configure_presets(
            self.presets.comfort_temperature, eco_temperature
        )

    async def async_configure_temperature_offset(
        self, temperature_offset: float
    ) -> Status:
        """Sets the thermostat's temperature offset."""

        return await self._async_write_command_with_status_response(
            OffsetConfigureCommand(offset=temperature_offset)
        )

    async def async_set_mode(self, operation_mode: OperationMode) -> Status:
        """Set new operation mode."""

        command: ModeSetCommand

        match operation_mode:
            case OperationMode.AUTO:
                command = ModeSetCommand(mode=OperationMode.AUTO)
            case OperationMode.MANUAL:
                command = ModeSetCommand(
                    mode=OperationMode.MANUAL
                    | Eq3Temperature.encode(self.status.target_temperature)
                )
            case OperationMode.OFF:
                command = ModeSetCommand(
                    mode=OperationMode.MANUAL | Eq3Temperature.encode(EQ3BT_OFF_TEMP)
                )
            case OperationMode.ON:
                command = ModeSetCommand(
                    mode=OperationMode.MANUAL | Eq3Temperature.encode(EQ3BT_ON_TEMP)
                )

        return await self._async_write_command_with_status_response(command)

    async def async_set_away(
        self,
        away_until: datetime,
        temperature: float,
    ) -> Status:
        """Set away mode."""

        return await self._async_write_command_with_status_response(
            AwaySetCommand(
                mode=OperationMode.AWAY | Eq3Temperature.encode(temperature),
                away_until=away_until,
            )
        )

    async def async_set_temperature(self, temperature: float) -> Status:
        """Set new target temperature."""

        if temperature == EQ3BT_OFF_TEMP:
            return await self.async_set_mode(OperationMode.OFF)

        if temperature == EQ3BT_ON_TEMP:
            return await self.async_set_mode(OperationMode.ON)

        return await self._async_write_command_with_status_response(
            TemperatureSetCommand(temperature=temperature)
        )

    async def async_set_preset(self, preset: Eq3Preset) -> Status:
        """Sets the thermostat to the given preset."""

        command: ComfortSetCommand | EcoSetCommand

        match preset:
            case Eq3Preset.COMFORT:
                command = ComfortSetCommand()
            case Eq3Preset.ECO:
                command = EcoSetCommand()

        return await self._async_write_command_with_status_response(command)

    async def async_set_boost(self, enable: bool) -> Status:
        """Sets boost mode."""

        return await self._async_write_command_with_status_response(
            BoostSetCommand(enable=enable)
        )

    async def async_set_locked(self, enable: bool) -> Status:
        """Locks or unlocks the thermostat."""

        return await self._async_write_command_with_status_response(
            LockSetCommand(enable=enable)
        )

    async def async_delete_schedule(
        self, week_days: list[WeekDay] | WeekDay | None = None
    ) -> Schedule:
        """Deletes the schedule."""

        week_days = (
            [week_days]
            if isinstance(week_days, WeekDay)
            else week_days or list(WeekDay)
        )

        return await self._async_write_commands_with_schedule_response(
            [ScheduleSetCommand(day=week_day, hours=[]) for week_day in week_days]
        )

    async def async_set_schedule(self, schedule: Schedule) -> Schedule:
        """Sets the schedule."""

        return await self._async_write_commands_with_schedule_response(
            [
                ScheduleSetCommand(
                    day=schedule_day.week_day,
                    hours=[
                        ScheduleHourStruct(
                            target_temp=schedule_hour.target_temperature,
                            next_change_at=schedule_hour.next_change_at,
                        )
                        for schedule_hour in schedule_day.schedule_hours
                    ],
                )
                for schedule_day in schedule.schedule_days
            ]
        )

    ### Internal ###

    async def __aenter__(self) -> Self:
        await self.async_connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        await self.async_disconnect()

    async def _async_write_command_with_device_data_response(
        self, command: Eq3Struct
    ) -> DeviceData:
        """Write a command to the thermostat and wait for a device data response."""

        if self._device_data_future is not None:
            raise Eq3AlreadyAwaitingResponseException(
                "Already awaiting a device data command response"
            )

        self._device_data_future = asyncio.Future()

        await self._async_write_command(command)

        try:
            device_data = await asyncio.wait_for(
                self._device_data_future, self._command_timeout
            )
        except TimeoutError as ex:
            raise Eq3TimeoutException("Timeout during device data command") from ex
        finally:
            self._device_data_future = None

        return device_data

    async def _async_write_command_with_status_response(
        self, command: Eq3Struct
    ) -> Status:
        """Write a command to the thermostat and wait for a status response."""

        if self._status_future is not None:
            raise Eq3AlreadyAwaitingResponseException(
                "Already awaiting a status command response"
            )

        self._status_future = asyncio.Future()

        await self._async_write_command(command)

        try:
            status = await asyncio.wait_for(self._status_future, self._command_timeout)
        except TimeoutError as ex:
            raise Eq3TimeoutException("Timeout during status command") from ex
        finally:
            self._status_future = None

        return status

    async def _async_write_commands_with_schedule_response(
        self, commands: list[Eq3Struct]
    ) -> Schedule:
        """Write commands to the thermostat and wait for a schedule response."""

        if self._schedule_future is not None:
            raise Eq3AlreadyAwaitingResponseException(
                "Already awaiting a schedule command response"
            )

        self._schedule_future = asyncio.Future()
        self._schedule_future_counter = len(commands)

        for command in commands:
            await self._async_write_command(command)

        try:
            schedule = await asyncio.wait_for(
                self._schedule_future, self._command_timeout
            )
        except TimeoutError as ex:
            raise Eq3TimeoutException("Timeout during schedule command") from ex
        finally:
            self._schedule_future = None

        return schedule

    async def _async_write_command(self, command: Eq3Struct) -> None:
        """Write a command to the thermostat."""

        if not self._conn.is_connected:
            raise Eq3ConnectionException("Not connected")

        data = command.to_bytes()

        async with self._lock:
            try:
                await asyncio.wait_for(
                    self._conn.write_gatt_char(Eq3Characteristic.WRITE, data),
                    self._command_timeout,
                )
            except BleakError as ex:
                raise Eq3CommandException("Error during write") from ex
            except TimeoutError as ex:
                raise Eq3TimeoutException("Timeout during write") from ex

    def _on_disconnected(self, client: BleakClient) -> None:
        """Handle disconnection from the thermostat."""

        asyncio.create_task(self.trigger_event(Eq3Event.DISCONNECTED))

    async def _on_message_received(
        self, handle: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle received messages from the thermostat."""

        data_bytes = bytes(data)
        command = DataclassStruct(Eq3Message).parse(data_bytes)

        match command.cmd:
            case Command.ID_RETURN:
                device_data = DeviceData.from_bytes(data_bytes)
                return await self._on_device_data_received(device_data)
            case Command.INFO_RETURN:
                if not command.is_status_command:
                    return

                status = Status.from_bytes(data_bytes)
                return await self._on_status_received(status)
            case Command.SCHEDULE_RETURN:
                schedule = Schedule.from_bytes(data_bytes)
                return await self._on_schedule_received(schedule)
            case _:
                raise Eq3InternalException(f"Unknown command: {command}")

    async def _on_device_data_received(self, device_data: DeviceData) -> None:
        """Handle received device data."""

        self._last_device_data = device_data

        if self._device_data_future is not None:
            return self._device_data_future.set_result(device_data)

        await self.trigger_event(Eq3Event.DEVICE_DATA_RECEIVED, device_data=device_data)

    async def _on_status_received(self, status: Status) -> None:
        """Handle received status."""

        self._last_status = status

        if self._status_future is not None:
            return self._status_future.set_result(status)

        await self.trigger_event(Eq3Event.STATUS_RECEIVED, status=status)

    async def _on_schedule_received(self, schedule: Schedule) -> None:
        """Handle received schedule."""

        if self._last_schedule is None:
            self._last_schedule = schedule

        self._last_schedule.merge(schedule)

        if self._schedule_future is not None:
            self._schedule_future_counter -= 1

            if self._schedule_future_counter == 0:
                return self._schedule_future.set_result(self._last_schedule)

        else:
            await self.trigger_event(
                Eq3Event.SCHEDULE_RECEIVED, schedule=self._last_schedule
            )

    ### Callbacks ###

    @overload
    def register_callback(
        self,
        event: Literal[Eq3Event.DISCONNECTED],
        callback: Union[Callable[[], None], Callable[[], Awaitable[None]]],
    ) -> None: ...

    @overload
    def register_callback(
        self,
        event: Union[Literal[Eq3Event.CONNECTED]],
        callback: Union[
            Callable[[DeviceData, Status, Schedule], None],
            Callable[[DeviceData, Status, Schedule], Awaitable[None]],
        ],
    ) -> None: ...

    @overload
    def register_callback(
        self,
        event: Literal[Eq3Event.DEVICE_DATA_RECEIVED],
        callback: Union[
            Callable[[DeviceData], None], Callable[[DeviceData], Awaitable[None]]
        ],
    ) -> None: ...

    @overload
    def register_callback(
        self,
        event: Literal[Eq3Event.STATUS_RECEIVED],
        callback: Union[Callable[[Status], None], Callable[[Status], Awaitable[None]]],
    ) -> None: ...

    @overload
    def register_callback(
        self,
        event: Literal[Eq3Event.SCHEDULE_RECEIVED],
        callback: Union[
            Callable[[Schedule], None], Callable[[Schedule], Awaitable[None]]
        ],
    ) -> None: ...

    def register_callback(
        self,
        event: Eq3Event,
        callback: Union[Callable[..., None], Callable[..., Awaitable[None]]],
    ) -> None:
        """Register a callback for a specific event."""

        if callback in self._callbacks[event]:
            return

        self._callbacks[event].append(callback)

    def unregister_callback(
        self,
        event: Eq3Event,
        callback: Union[Callable[..., None], Callable[..., Awaitable[None]]],
    ) -> None:
        """Unregister a callback for a specific event."""

        if callback not in self._callbacks[event]:
            return

        self._callbacks[event].remove(callback)

    @overload
    async def trigger_event(self, event: Literal[Eq3Event.DISCONNECTED]) -> None: ...

    @overload
    async def trigger_event(
        self,
        event: Literal[Eq3Event.CONNECTED],
        *,
        device_data: DeviceData,
        status: Status,
        schedule: Schedule,
    ) -> None: ...

    @overload
    async def trigger_event(
        self,
        event: Literal[Eq3Event.DEVICE_DATA_RECEIVED],
        *,
        device_data: DeviceData,
    ) -> None: ...

    @overload
    async def trigger_event(
        self,
        event: Literal[Eq3Event.STATUS_RECEIVED],
        *,
        status: Status,
    ) -> None: ...

    @overload
    async def trigger_event(
        self,
        event: Literal[Eq3Event.SCHEDULE_RECEIVED],
        *,
        schedule: Schedule,
    ) -> None: ...

    async def trigger_event(
        self,
        event: Eq3Event,
        *,
        device_data: DeviceData | None = None,
        status: Status | None = None,
        schedule: Schedule | None = None,
    ) -> None:
        """Call the callbacks for a specific event."""

        async_callbacks = [
            callback
            for callback in self._callbacks[event]
            if asyncio.iscoroutinefunction(callback)
        ]
        sync_callbacks = [
            callback
            for callback in self._callbacks[event]
            if not asyncio.iscoroutinefunction(callback)
        ]

        args: (
            tuple[DeviceData, Status, Schedule]
            | tuple[DeviceData]
            | tuple[Status]
            | tuple[Schedule]
            | tuple[()]
        )

        match event:
            case Eq3Event.DISCONNECTED:
                args = ()
            case Eq3Event.CONNECTED:
                if device_data is None or status is None or schedule is None:
                    raise Eq3InternalException(
                        "device_data, status, and schedule must not be None for CONNECTED event"
                    )
                args = (device_data, status, schedule)
            case Eq3Event.DEVICE_DATA_RECEIVED:
                if device_data is None:
                    raise Eq3InternalException(
                        "device_data must not be None for DEVICE_DATA_RECEIVED event"
                    )
                args = (device_data,)
            case Eq3Event.STATUS_RECEIVED:
                if status is None:
                    raise Eq3InternalException(
                        "status must not be None for STATUS_RECEIVED event"
                    )
                args = (status,)
            case Eq3Event.SCHEDULE_RECEIVED:
                if schedule is None:
                    raise Eq3InternalException(
                        "schedule must not be None for SCHEDULE_RECEIVED event"
                    )
                args = (schedule,)

        await asyncio.gather(*[callback(*args) for callback in async_callbacks])

        for callback in sync_callbacks:
            callback(*args)



"""Constants for the eq3btsmart library."""

from enum import IntEnum, StrEnum, auto

from construct_typed import EnumBase, FlagsEnumBase

EQ3BT_AWAY_TEMP = 12.0
EQ3BT_MIN_TEMP = 5.0
EQ3BT_MAX_TEMP = 29.5
EQ3BT_OFF_TEMP = 4.5
EQ3BT_ON_TEMP = 30.0
EQ3BT_MIN_OFFSET = -3.5
EQ3BT_MAX_OFFSET = 3.5

DEFAULT_CONNECTION_TIMEOUT = 10
DEFAULT_COMMAND_TIMEOUT = 5


class Eq3Characteristic(StrEnum):
    """Characteristics."""

    WRITE = "3fa4585a-ce4a-3bad-db4b-b8df8179ea09"
    NOTIFY = "d0e8434d-cd29-0996-af41-6c90f4e0eb2a"


class Command(EnumBase):
    ID_GET = 0x00
    ID_RETURN = 0x01
    INFO_RETURN = 0x02
    INFO_GET = 0x03
    SCHEDULE_SET = 0x10
    COMFORT_ECO_CONFIGURE = 0x11
    OFFSET_CONFIGURE = 0x13
    WINDOW_OPEN_CONFIGURE = 0x14
    SCHEDULE_GET = 0x20
    SCHEDULE_RETURN = 0x21
    MODE_SET = 0x40
    TEMPERATURE_SET = 0x41
    COMFORT_SET = 0x43
    ECO_SET = 0x44
    BOOST_SET = 0x45
    LOCK_SET = 0x80


class WeekDay(EnumBase):
    """Weekdays."""

    SATURDAY = 0
    SUNDAY = 1
    MONDAY = 2
    TUESDAY = 3
    WEDNESDAY = 4
    THURSDAY = 5
    FRIDAY = 6

    @classmethod
    def from_index(cls, index: int) -> "WeekDay":
        """Return weekday from index."""

        adjusted_index = index + 2 if index < 5 else index - 5
        return cls(adjusted_index)


class OperationMode(IntEnum):
    """Operation modes."""

    AUTO = 0x00
    MANUAL = 0x40
    OFF = 0x49
    ON = 0x7B
    AWAY = 0x80


class StatusFlags(FlagsEnumBase):
    """Status flags."""

    AUTO = 0x00  # always True, doesnt affect building
    MANUAL = 0x01
    AWAY = 0x02
    BOOST = 0x04
    DST = 0x08
    WINDOW = 0x10
    LOCKED = 0x20
    UNKNOWN = 0x40
    LOW_BATTERY = 0x80


class Eq3Preset(IntEnum):
    """Preset modes."""

    COMFORT = 0
    ECO = 1


class Eq3Event(StrEnum):
    """Event types."""

    CONNECTED = auto()
    DISCONNECTED = auto()
    DEVICE_DATA_RECEIVED = auto()
    STATUS_RECEIVED = auto()
    SCHEDULE_RECEIVED = auto()



from eq3btsmart.models.device_data import DeviceData
from eq3btsmart.structures import DeviceDataStruct


def test_device_data_from_bytes() -> None:
    mock_struct = DeviceDataStruct(
        version=1, serial="OEQ1750973", unknown_1=0, unknown_2=0, unknown_3=0
    )
    mock_bytes = mock_struct.to_bytes()

    device_data = DeviceData.from_bytes(mock_bytes)

    assert device_data.firmware_version == 1
    assert device_data.device_serial == "OEQ1750973"



from datetime import time

from eq3btsmart.const import WeekDay
from eq3btsmart.models import Schedule, ScheduleDay
from eq3btsmart.models.schedule_hour import ScheduleHour
from eq3btsmart.structures import ScheduleDayStruct, ScheduleHourStruct


def test_schedule_initialization() -> None:
    schedule = Schedule()
    assert schedule.schedule_days == [], "Initial schedule_days should be empty"


def test_schedule_merge() -> None:
    schedule1 = Schedule(
        schedule_days=[
            ScheduleDay(
                week_day=WeekDay.MONDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=1, minute=0),
                        target_temperature=20,
                    ),
                ],
            )
        ]
    )
    schedule2 = Schedule(
        schedule_days=[
            ScheduleDay(
                week_day=WeekDay.TUESDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=1, minute=0),
                        target_temperature=20,
                    ),
                ],
            )
        ]
    )
    schedule3 = Schedule(
        schedule_days=[
            ScheduleDay(
                week_day=WeekDay.MONDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=2, minute=0),
                        target_temperature=21,
                    ),
                ],
            )
        ]
    )

    schedule1.merge(schedule2)
    schedule1.merge(schedule3)

    assert schedule1 == Schedule(
        schedule_days=[
            ScheduleDay(
                week_day=WeekDay.MONDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=2, minute=0),
                        target_temperature=21,
                    ),
                ],
            ),
            ScheduleDay(
                week_day=WeekDay.TUESDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=1, minute=0),
                        target_temperature=20,
                    ),
                ],
            ),
        ]
    )


def test_schedule_from_bytes() -> None:
    schedule_day_structs = [
        ScheduleDayStruct(
            day=WeekDay.MONDAY,
            hours=[
                ScheduleHourStruct(
                    target_temp=20,
                    next_change_at=time(hour=1, minute=0),
                ),
            ],
        )
    ]
    schedule_days_bytes = [
        schedule_day_struct.to_bytes() for schedule_day_struct in schedule_day_structs
    ]
    schedule_day_structs_from_bytes = [
        ScheduleDay.from_bytes(schedule_day_bytes)
        for schedule_day_bytes in schedule_days_bytes
    ]

    assert [
        ScheduleDay(
            week_day=WeekDay.MONDAY,
            schedule_hours=[
                ScheduleHour(
                    next_change_at=time(hour=1, minute=0),
                    target_temperature=20,
                ),
            ],
        )
    ] == schedule_day_structs_from_bytes


def test_schedule_delete_day() -> None:
    schedule = Schedule(
        schedule_days=[
            ScheduleDay(
                week_day=WeekDay.MONDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=1, minute=0),
                        target_temperature=20,
                    ),
                ],
            ),
            ScheduleDay(
                week_day=WeekDay.TUESDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=1, minute=0),
                        target_temperature=20,
                    ),
                ],
            ),
        ]
    )
    schedule.delete_day(WeekDay.MONDAY)

    assert schedule == Schedule(
        schedule_days=[
            ScheduleDay(
                week_day=WeekDay.TUESDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=1, minute=0),
                        target_temperature=20,
                    ),
                ],
            ),
        ]
    )


def test_eq() -> None:
    schedule1 = Schedule(
        schedule_days=[
            ScheduleDay(
                week_day=WeekDay.MONDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=1, minute=0),
                        target_temperature=20,
                    ),
                ],
            )
        ]
    )
    schedule2 = Schedule(
        schedule_days=[
            ScheduleDay(
                week_day=WeekDay.MONDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=1, minute=0),
                        target_temperature=20,
                    ),
                ],
            )
        ]
    )
    schedule3 = Schedule(
        schedule_days=[
            ScheduleDay(
                week_day=WeekDay.MONDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=1, minute=0),
                        target_temperature=20,
                    ),
                ],
            ),
            ScheduleDay(
                week_day=WeekDay.TUESDAY,
                schedule_hours=[],
            ),
        ]
    )
    schedule4 = Schedule(
        schedule_days=[
            ScheduleDay(
                week_day=WeekDay.MONDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=1, minute=0),
                        target_temperature=20,
                    ),
                ],
            ),
            ScheduleDay(
                week_day=WeekDay.TUESDAY,
                schedule_hours=[
                    ScheduleHour(
                        next_change_at=time(hour=1, minute=0),
                        target_temperature=20,
                    ),
                ],
            ),
        ]
    )

    assert schedule1 == schedule2
    assert not schedule1 == "foo"
    assert not schedule1 == Schedule(schedule_days=[])
    assert schedule1 == schedule3
    assert schedule3 != schedule4



from construct_typed import Adapter, Context


class Eq3Serial(Adapter[bytes, bytes, str, str]):
    """Adapter to encode and decode serial id."""

    def _encode(self, obj: str, ctx: Context, path: str) -> bytes:
        return self.encode(obj)

    def _decode(self, obj: bytes, ctx: Context, path: str) -> str:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: str) -> bytes:
        return bytes(char + 0x30 for char in value.encode())

    @classmethod
    def decode(cls, value: bytes) -> str:
        return bytes(char - 0x30 for char in value).decode()



from datetime import timedelta

from construct_typed import Adapter, Context

from eq3btsmart.exceptions import Eq3InvalidDataException


class Eq3Duration(Adapter[int, int, timedelta, timedelta]):
    """Adapter to encode and decode duration data."""

    def _encode(self, obj: timedelta, ctx: Context, path: str) -> int:
        return self.encode(obj)

    def _decode(self, obj: int, ctx: Context, path: str) -> timedelta:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: timedelta) -> int:
        if value.seconds < 0 or value.seconds > 3600.0:
            raise Eq3InvalidDataException(
                "Window open time must be between 0 and 60 minutes in intervals of 5 minutes."
            )

        return int(value.seconds / 300.0)

    @classmethod
    def decode(cls, value: int) -> timedelta:
        return timedelta(minutes=float(value * 5.0))



from eq3btsmart.const import OperationMode, StatusFlags
from eq3btsmart.models import Status
from eq3btsmart.structures import StatusStruct


def test_status_from_device() -> None:
    mock_struct = StatusStruct(
        valve=50,
        target_temp=21,
        mode=StatusFlags.MANUAL,
    )

    status = Status.from_struct(mock_struct)

    assert status.valve == 50
    assert status.target_temperature == 21


def test_operation_mode_property() -> None:
    status = Status(
        valve=5,
        target_temperature=4.5,
        _operation_mode=OperationMode.ON,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )
    assert status.operation_mode == OperationMode.OFF

    status = Status(
        valve=5,
        target_temperature=24,
        _operation_mode=OperationMode.AUTO,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )
    assert status.operation_mode == OperationMode.AUTO

    status = Status(
        valve=5,
        target_temperature=30,
        _operation_mode=OperationMode.AUTO,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )
    assert status.operation_mode == OperationMode.ON

    status = Status(
        valve=5,
        target_temperature=20,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )
    assert status.operation_mode == OperationMode.MANUAL


def test_status_from_bytes() -> None:
    mock_struct = StatusStruct(
        valve=50,
        target_temp=21,
        mode=StatusFlags.MANUAL,
    )
    mock_bytes = mock_struct.to_bytes()

    status = Status.from_bytes(mock_bytes)

    assert status.valve == 50
    assert status.target_temperature == 21


def test_valve_temperature() -> None:
    status = Status(
        valve=0,
        target_temperature=20,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )
    assert status.valve_temperature == 20

    status = Status(
        valve=50,
        target_temperature=20,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )
    assert status.valve_temperature == 19

    status = Status(
        valve=100,
        target_temperature=20,
        _operation_mode=OperationMode.MANUAL,
        is_away=False,
        is_boost=False,
        is_dst=False,
        is_window_open=False,
        is_locked=False,
        is_low_battery=False,
        away_until=None,
        presets=None,
    )
    assert status.valve_temperature == 18



"""Structures for the eQ-3 Bluetooth Smart Thermostat."""

from dataclasses import dataclass
from datetime import datetime, time, timedelta
from typing import Self

from construct import (
    Bytes,
    Const,
    Flag,
    GreedyBytes,
    GreedyRange,
    Int8ub,
    Optional,
)
from construct_typed import DataclassMixin, DataclassStruct, TEnum, TFlagsEnum, csfield

from eq3btsmart.adapter.eq3_away_time import Eq3AwayTime
from eq3btsmart.adapter.eq3_duration import Eq3Duration
from eq3btsmart.adapter.eq3_schedule_time import Eq3ScheduleTime
from eq3btsmart.adapter.eq3_serial import Eq3Serial
from eq3btsmart.adapter.eq3_temperature import Eq3Temperature
from eq3btsmart.adapter.eq3_temperature_offset import Eq3TemperatureOffset
from eq3btsmart.adapter.eq3_time import Eq3Time
from eq3btsmart.const import (
    Command,
    StatusFlags,
    WeekDay,
)


class Eq3Struct(DataclassMixin):
    """Structure for eQ-3 data."""

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        """Convert the data to a structure."""

        return DataclassStruct(cls).parse(data)

    def to_bytes(self) -> bytes:
        """Convert the structure to bytes."""

        return DataclassStruct(self.__class__).build(self)


@dataclass
class PresetsStruct(Eq3Struct):
    """Structure for presets data."""

    window_open_temp: float = csfield(Eq3Temperature(Int8ub))
    window_open_time: timedelta = csfield(Eq3Duration(Int8ub))
    comfort_temp: float = csfield(Eq3Temperature(Int8ub))
    eco_temp: float = csfield(Eq3Temperature(Int8ub))
    offset: float = csfield(Eq3TemperatureOffset(Int8ub))


@dataclass
class StatusStruct(Eq3Struct):
    """Structure for status data."""

    cmd: int = csfield(Const(Command.INFO_RETURN, Int8ub))
    const_1: int = csfield(Const(0x01, Int8ub))
    mode: StatusFlags = csfield(TFlagsEnum(Int8ub, StatusFlags))
    valve: int = csfield(Int8ub)
    const_2: int = csfield(Const(0x04, Int8ub))
    target_temp: float = csfield(Eq3Temperature(Int8ub))
    away: datetime | None = csfield(Optional(Eq3AwayTime(Bytes(4))))
    presets: PresetsStruct | None = csfield(Optional(DataclassStruct(PresetsStruct)))


@dataclass
class ScheduleHourStruct(Eq3Struct):
    """Structure for schedule entry data."""

    target_temp: float = csfield(Eq3Temperature(Int8ub))
    next_change_at: time = csfield(Eq3ScheduleTime(Int8ub))


@dataclass
class ScheduleDayStruct(Eq3Struct):
    """Structure for schedule data."""

    cmd: int = csfield(Const(Command.SCHEDULE_RETURN, Int8ub))
    day: WeekDay = csfield(TEnum(Int8ub, WeekDay))
    hours: list[ScheduleHourStruct] = csfield(
        GreedyRange(DataclassStruct(ScheduleHourStruct))
    )


@dataclass
class DeviceDataStruct(Eq3Struct):
    """Structure for device data."""

    cmd: int = csfield(Const(Command.ID_RETURN, Int8ub))
    version: int = csfield(Int8ub)
    unknown_1: int = csfield(Int8ub)
    unknown_2: int = csfield(Int8ub)
    serial: str = csfield(Eq3Serial(Bytes(10)))
    unknown_3: int = csfield(Int8ub)


@dataclass
class Eq3Message(Eq3Struct):
    """Structure for eQ-3 message."""

    cmd: int = csfield(Int8ub)
    is_status_command: bool = csfield(Flag)
    data: bytes = csfield(GreedyBytes)


@dataclass
class IdGetCommand(Eq3Struct):
    """Structure for ID get command."""

    cmd: int = csfield(Const(Command.ID_GET, Int8ub))


@dataclass
class InfoGetCommand(Eq3Struct):
    """Structure for info get command."""

    cmd: int = csfield(Const(Command.INFO_GET, Int8ub))
    time: datetime = csfield(Eq3Time(Bytes(6)))


@dataclass
class ComfortEcoConfigureCommand(Eq3Struct):
    """Structure for schedule get command."""

    cmd: int = csfield(Const(Command.COMFORT_ECO_CONFIGURE, Int8ub))
    comfort_temperature: float = csfield(Eq3Temperature(Int8ub))
    eco_temperature: float = csfield(Eq3Temperature(Int8ub))


@dataclass
class OffsetConfigureCommand(Eq3Struct):
    """Structure for offset configure command."""

    cmd: int = csfield(Const(Command.OFFSET_CONFIGURE, Int8ub))
    offset: float = csfield(Eq3TemperatureOffset(Int8ub))


@dataclass
class WindowOpenConfigureCommand(Eq3Struct):
    """Structure for window open configure command."""

    cmd: int = csfield(Const(Command.WINDOW_OPEN_CONFIGURE, Int8ub))
    window_open_temperature: float = csfield(Eq3Temperature(Int8ub))
    window_open_time: timedelta = csfield(Eq3Duration(Int8ub))


@dataclass
class ScheduleGetCommand(Eq3Struct):
    """Structure for schedule get command."""

    cmd: int = csfield(Const(Command.SCHEDULE_GET, Int8ub))
    day: WeekDay = csfield(TEnum(Int8ub, WeekDay))


@dataclass
class ModeSetCommand(Eq3Struct):
    """Structure for mode set command."""

    cmd: int = csfield(Const(Command.MODE_SET, Int8ub))
    mode: int = csfield(Int8ub)


@dataclass
class AwaySetCommand(ModeSetCommand):
    """Structure for away set command."""

    away_until: datetime = csfield(Eq3AwayTime(Bytes(4)))


@dataclass
class TemperatureSetCommand(Eq3Struct):
    """Structure for temperature set command."""

    cmd: int = csfield(Const(Command.TEMPERATURE_SET, Int8ub))
    temperature: float = csfield(Eq3Temperature(Int8ub))


@dataclass
class ScheduleSetCommand(Eq3Struct):
    """Structure for schedule set command."""

    cmd: int = csfield(Const(Command.SCHEDULE_SET, Int8ub))
    day: WeekDay = csfield(TEnum(Int8ub, WeekDay))
    hours: list[ScheduleHourStruct] = csfield(
        GreedyRange(DataclassStruct(ScheduleHourStruct))
    )


@dataclass
class ComfortSetCommand(Eq3Struct):
    """Structure for comfort set command."""

    cmd: int = csfield(Const(Command.COMFORT_SET, Int8ub))


@dataclass
class EcoSetCommand(Eq3Struct):
    """Structure for eco set command."""

    cmd: int = csfield(Const(Command.ECO_SET, Int8ub))


@dataclass
class BoostSetCommand(Eq3Struct):
    """Structure for boost set command."""

    cmd: int = csfield(Const(Command.BOOST_SET, Int8ub))
    enable: bool = csfield(Flag)


@dataclass
class LockSetCommand(Eq3Struct):
    """Structure for lock set command."""

    cmd: int = csfield(Const(Command.LOCK_SET, Int8ub))
    enable: bool = csfield(Flag)
